/*
 *  Copyright (C) Chaperon. All rights reserved.
 *  -------------------------------------------------------------------------
 *  This software is published under the terms of the Apache Software License
 *  version 1.1, a copy of which has been included  with this distribution in
 *  the LICENSE file.
 */

package net.sourceforge.chaperon.parser;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;

import org.xml.sax.ContentHandler;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.AttributesImpl;

import net.sourceforge.chaperon.helpers.Decoder;

/**
 * A automate for a regex expression
 *
 * @author Stephan Michels
 * @version CVS $Id: RegexAutomate.java,v 1.5 2002/05/06 16:22:56 benedikta Exp $
 */
public class RegexAutomate implements Serializable
{

	/** A simple transition without matching */
	public final static int TYPE_NOMATCH = 0;

	/** If the character should match */
	public final static int TYPE_MATCH = 1;

	/** If the characher shouldn't match */
	public final static int TYPE_NEGMATCH = 2;

	/** Match all characters */
	public final static int TYPE_MATCHANY = 3;

	/** Begin of line */
	public final static int TYPE_BOL = 4;

	/** End of line */
	public final static int TYPE_EOL = 5;

  /** End of file */
  public final static int TYPE_EOF = 6;

	private transient int[] _types = new int[0];

	private transient String[] _chars = new String[0];
	private transient int[][] _transitions = new int[0][0];

	// Count of states
	private transient int _count = 0;
	// Size of arrays
	private transient int _size = 0;

	// Accepted state
	private transient int _acceptstate = -1;

	private static final long serialVersionUID = 1246342009422281917L;

	/**
	 * Constructs a empty automate
	 */
	public RegexAutomate()
	{
	}

	/**
	 * Adds a state to the automate
	 * 
	 * @param chars Characters to compare
	 * @param type Type of state
	 * @param transitions The destination states, if the match was successful
	 *
	 * @return Index of the added state
	 */
	public int addState(String chars, int type, int[] transitions)
	{
		ensureCapacity(_count + 1);
		_chars[_count] = chars;
		_types[_count] = type;
		_transitions[_count] = transitions;
		_count++;
		return _count - 1;
	}

	/*
	 * Adds a state to the automate.
	 * The destination states were not defined
	 *
	 * @param chars Characters to compare
	 * @param type Type of state
	 *
	 * @return Index of the added state
	 */
	public int addState(String chars, int type)
	{
		ensureCapacity(_count + 1);
		_chars[_count] = chars;
		_types[_count] = type;
		_transitions[_count] = new int[0];
		_count++;
		return _count - 1;
	}

	/**
	 * Adds a state to the automate.
	 *
	 * @param type Type of state
	 * @param transitions The destination states
	 *
	 * @return Index of the added state
	 */
	public int addState(int type, int[] transitions)
	{
		ensureCapacity(_count + 1);
		_chars[_count] = null;
		_types[_count] = type;
		_transitions[_count] = transitions;
		_count++;
		return _count - 1;
	}

	/**
	 * Adds a state to the automate.
	 *
	 * @return Index of the added state
	 */
	public int addState()
	{
		ensureCapacity(_count + 1);
		_chars[_count] = null;
		_types[_count] = TYPE_NOMATCH;
		_transitions[_count] = new int[0];
		_count++;
		return _count - 1;
	}

	/**
	 * Adds a no comparing state to the automate.
	 *
	 * @param transitions The destination states
	 *
	 * @return Index of the added state
	 */
	public int addState(int[] transitions)
	{
		ensureCapacity(_count + 1);
		_chars[_count] = null;
		_types[_count] = TYPE_NOMATCH;
		_transitions[_count] = transitions;
		_count++;
		return _count - 1;
	}

	/**
	 * Add a transition to a state
	 *
	 * @param state Index of the state
	 * @param nextstate Destination state
	 */
	public void addTransition(int state, int nextstate)
	{
		if ((state < 0) || (state >= _count))
			throw new IndexOutOfBoundsException();

		int[] newtransitions = new int[_transitions[state].length + 1];

		System.arraycopy(_transitions[state], 0, newtransitions, 0,
										 _transitions[state].length);
		newtransitions[_transitions[state].length] = nextstate;
		_transitions[state] = newtransitions;
	}

	/**
	 * Removes a state from the automate
	 *
	 * @param index Index of the state
	 */
	public void removeState(int index)
	{
		int i, j;

		for (i = 0; i < index; i++)
		{
			for (j = 0; j < _transitions[i].length; j++)
				if (_transitions[i][j] == index)
					_transitions[i][j] = -1;
				else if (_transitions[i][j] > index)
					_transitions[i][j] -= 1;
		}

		for (i = index + 1; i < _count; i++)
		{
			_chars[i - 1] = _chars[i];
			_types[i - 1] = _types[i];

			for (j = 0; j < _transitions[i].length; j++)
				if (_transitions[i][j] == index)
					_transitions[i - 1][j] = -1;
				else if (_transitions[i][j] > index)
					_transitions[i - 1][j] = _transitions[i][j] - 1;
				else
					_transitions[i - 1][j] = _transitions[i][j];
		}
		_count--;
	}

	/**
	 * Replace a state
	 *
	 * @param index Inde of the state
	 * @param chars Characters to compare
	 * @param type Type of state
	 * @param transitions The destination states, if the match was successful
	 */
	public void setState(int index, String chars, int type, int[] transitions)
	{
		ensureCapacity(index + 1);
		_chars[index] = chars;
		_types[index] = type;
		_transitions[index] = transitions;
	}

	/**
	 * Sets the end state. If the automate reaches this
	 * state, the automate was successful
	 *
	 * @param acceptstate Index of the accepted state
	 */
	public void setAcceptState(int acceptstate)
	{
		if ((acceptstate < 0) || (acceptstate > _count))
			throw new IllegalArgumentException();
		_acceptstate = acceptstate;
	}

	/**
	 * Returns the index of the end state
	 *
	 * @return Index of the end state
	 */
	public int getAcceptState()
	{
		return _acceptstate;
	}

	/**
	 * Test if the state is the end state
	 *
	 * @param state Index of the state
	 *
	 * @return True, if the state is the end state
	 */
	public boolean isAcceptState(int state)
	{
		return _acceptstate == state;
	}

	/**
	 * Return the to comparing characters
	 *
	 * @param state Index of the state
	 *
	 * @return To comparing characters
	 */
	public String getChars(int state)
	{
		if ((state < 0) || (state >= _count))
			throw new IndexOutOfBoundsException();
		return _chars[state];
	}

	/**
	 * Returns the type of the state
	 *
	 * @param state Index of the state
	 *
	 * @return Type of state
	 */
	public int getType(int state)
	{
		if ((state < 0) || (state >= _count))
			throw new IndexOutOfBoundsException();
		return _types[state];
	}

	/**
	 * Returns the destinations of the transitions
	 *
	 * @param state Index of the state
	 *
	 * @return Destinations of the transitions
	 */
	public int[] getTransitions(int state)
	{
		if ((state < 0) || (state >= _count))
			throw new IndexOutOfBoundsException();
		return _transitions[state];
	}

	/**
	 * Test the state
	 *
	 * @param state Index of the state    
	 * @param text Text, which should compared
	 * @param position Position in the text
	 *
	 * @return True, if the compare is successful
	 */
	public boolean match(int state, char[] text, int position)
	{
		switch (_types[state])
		{
		case TYPE_NOMATCH :
			return true;

		case TYPE_MATCH :
			if (position < text.length)
				return _chars[state].indexOf(text[position]) >= 0;
			else
				return false;
		case TYPE_NEGMATCH :
			if (position < text.length)
				return _chars[state].indexOf(text[position]) < 0;
			else
				return false;
		case TYPE_MATCHANY :
			if (position < text.length)
				return ("\n\r".indexOf(text[position])) < 0;
			else
				return false;
		case TYPE_BOL :
			if (position == 0)
				return true;
			else if ((position == 1)
							 && (((text[position - 1] == '\n') && (text[position] != '\r'))
									 || (text[position - 1] == '\r')))
				return true;
			else if ((text[position - 1] == '\r')
							 || ((text[position - 1] == '\n') && (text[position] != '\r')))
				return true;
			else
				return false;
		case TYPE_EOL :
			if (position >= text.length)
				return true;
			else if (((position + 1) == text.length) && (text[position] != '\r')
							 && (text[position] != '\n'))
				return true;
			else if ((text[position] != '\r') && (text[position] != '\n')
							 && ((text[position + 1] == '\n')
									 || (text[position + 1] == '\r')))
				return true;
			else
				return false;
    case TYPE_EOF :
      if (position >= text.length)
        return true;
      else
        return false;
		}
		return false;
	}

	/**
	 * If the state consumes characters. 
	 *
	 * @param state Index of the state
	 *
	 * @return True, if the state consumes character for matching
	 */
	public boolean consumeCharacter(int state)
	{
		switch (_types[state])
		{
		case TYPE_NOMATCH :
			return false;

		case TYPE_MATCH :
			return true;

		case TYPE_NEGMATCH :
			return true;

		case TYPE_MATCHANY :
			return true;

		case TYPE_BOL :
			return false;

		case TYPE_EOL :
			return false;

    case TYPE_EOF :
      return false;
		}
		return false;
	}

	/**
	 * Returns the count of states
	 *
	 * @return Count of states
	 */
	public int getSize()
	{
		return _count;
	}

	/**
	 * Ensure the capacity for adding states
	 *
	 * @param minCapacity
	 */
	private void ensureCapacity(int minCapacity)
	{
		if (_size >= minCapacity)
			return;

		if (_size == 0)
		{
			_chars = new String[minCapacity];
			_types = new int[minCapacity];
			for (int i = 0; i < minCapacity; i++)
			{
				_chars[i] = null;
				_types[i] = TYPE_NOMATCH;
			}

			_transitions = new int[minCapacity][0];
		}

		int newCapacity = Math.max(_size + 10, minCapacity);

		String[] newchars = new String[newCapacity];
		int[] newtypes = new int[newCapacity];

		int[][] newtransitions = new int[newCapacity][];

		System.arraycopy(_chars, 0, newchars, 0, _size);
		System.arraycopy(_types, 0, newtypes, 0, _size);
		for (int i = 0; i < _size; i++)
		{
			newtransitions[i] = new int[_transitions[i].length];
			System.arraycopy(_transitions[i], 0, newtransitions[i], 0,
											 _transitions[i].length);
		}
		_chars = newchars;
		_types = newtypes;
		_transitions = newtransitions;
		_size = newCapacity;
	}

	/**
	 * Serialize the automate to a output stream
	 *
	 * @param out Output stream
	 *
	 * @throws IOException If an exception occurs, while writing the object
   *                     to the output stream.
	 */
	private void writeObject(ObjectOutputStream out) throws IOException
	{
		out.defaultWriteObject();
		int i, j;

		out.writeInt(_count);
		for (i = 0; i < _count; i++)
		{
			if (_chars[i] == null)
				out.writeUTF("");
			else
				out.writeUTF(_chars[i]);

			out.writeInt(_types[i]);

			out.writeInt(_transitions[i].length);
			for (j = 0; j < _transitions[i].length; j++)
				out.writeInt(_transitions[i][j]);
		}
		out.writeInt(_acceptstate);

		out.flush();
	}

	/**
	 * Read a automate from a input stream
	 *
	 * @param in InputStream
	 *
	 * @throws ClassNotFoundException If the class were not found
	 * @throws IOException If a error occurs, while reading the input stream
	 */
	private void readObject(ObjectInputStream in)
					throws IOException, ClassNotFoundException
	{
		in.defaultReadObject();

		int i, j, stringLength;
		StringBuffer buffer;

		int length = in.readInt();

		_chars = new String[length];
		_types = new int[length];
		_transitions = new int[length][];

		_size = length;
		_count = length;

		for (i = 0; i < length; i++)
		{
			_chars[i] = in.readUTF();

			_types[i] = in.readInt();

			_transitions[i] = new int[in.readInt()];
			for (j = 0; j < _transitions[i].length; j++)
				_transitions[i][j] = in.readInt();
		}
		_acceptstate = in.readInt();
	}

	/**
	 * Creates empty string for indenting
	 *
	 * @param n Count of character to indent
	 *
	 * @return Empty string
	 */
	private String distance(int n)
	{
		String string = new String();

		for (int i = 0; i < n; i++)
			string += " ";
		return string;
	}

	private static final String spaces = "                                                                                              ";

	/**
	 * Convert the automate to a string
	 *
	 * @return Content of the automate
	 */
	public String toString()
	{
		if ((_chars.length != _types.length)
						|| (_types.length != _transitions.length))
		{
			return "Automate is not valid. Lengths does not match.";
		}

		String[] chars = new String[_chars.length];
		int i;

		for (i = 0; i < _count; i++)
			switch (_types[i])
			{
			case TYPE_NOMATCH :
				chars[i] = " ";
			case TYPE_MATCH :
				chars[i] = "[" + Decoder.decode(_chars[i]) + "]";
			case TYPE_NEGMATCH :
				chars[i] = "[^" + Decoder.decode(_chars[i]) + "]";
			case TYPE_MATCHANY :
				chars[i] = ".";
			case TYPE_BOL :
				chars[i] = "^";
			case TYPE_EOL :
				chars[i] = "$";
      case TYPE_EOF :
        chars[i] = "$$";
			}

		String string = new String();
		String dummy;

		for (i = 0; i < _count; i++)
		{
			dummy = String.valueOf(i);
			string += " " + dummy
								+ spaces.substring(0, Math.max(0, chars[i].length()
																							 - dummy.length() + 1));
		}
		string += "\n";

		for (i = 0; i < _count; i++)
			string += " " + chars[i] + " ";
		string += "\n";

		int maxtransitions = 0;

		System.out.print("transitions = ");
		for (i = 0; i < _count; i++)
		{
			maxtransitions = Math.max(maxtransitions, _transitions[i].length);
			System.out.print(_transitions[i].length + " ");
		}
		System.out.println();
		for (int j = 0; j < maxtransitions; j++)
		{
			for (i = 0; i < _count; i++)
				if (j < _transitions[i].length)
				{
					dummy = String.valueOf(_transitions[i][j]);
					string += " " + dummy
										+ spaces.substring(0, Math.max(0, chars[i].length()
																									 - dummy.length() + 1));
				}
				else
					string += " "
										+ spaces.substring(0, Math.max(0, chars[i].length() + 1));
			string += "\n";
		}

		string += "Accepted state=" + _acceptstate;
		return string;

	}
}
